@@ -1,7 +1,9 @@
 #! /usr/bin/env python
 import future
 import os
+import codecs
 import sys
+import locale
 import logging
 from io import StringIO, open
 from . import csvhelpers
@@ -30,7 +32,7 @@ class CSVDedupe(csvhelpers.CSVCommand) :
                     raise self.parser.error("No input file or STDIN specified.")
             else:
                 try:
-                    self.input = open(self.configuration['input'], 'rU').read()
+                    self.input = open(self.configuration['input'], encoding='utf-8').read()
                 except IOError:
                     raise self.parser.error("Could not find the file %s" %
                                             (self.configuration['input'], ))
@@ -46,8 +48,8 @@ class CSVDedupe(csvhelpers.CSVCommand) :
             except KeyError:
                 raise self.parser.error("You must provide field_names")
         else :
-            self.field_names = [self.field_def['field'] 
-                                for field_def in self.field_definitions]
+            self.field_names = [field_def['field']
+                                for field_def in self.field_definition]
 
         self.destructive = self.configuration.get('destructive', False)
 
@@ -64,7 +66,7 @@ class CSVDedupe(csvhelpers.CSVCommand) :
         data_d = {}
         # import the specified CSV file
 
-        data_d = csvhelpers.readData(self.input, self.field_names)
+        data_d = csvhelpers.readData(self.input, self.field_names, delimiter=self.delimiter)
 
         logging.info('imported %d rows', len(data_d))
 
@@ -78,43 +80,34 @@ class CSVDedupe(csvhelpers.CSVCommand) :
 
         logging.info('using fields: %s' % [field['field']
                                            for field in self.field_definition])
-        # # Create a new deduper object and pass our data model to it.
-        deduper = dedupe.Dedupe(self.field_definition)
 
-        # Set up our data sample
-        logging.info('taking a sample of %d possible pairs', self.sample_size)
-        deduper.sample(data_d, self.sample_size)
+        # If --skip_training has been selected, and we have a settings cache still
+        # persisting from the last run, use it in this next run.
+        # __Note:__ if you want to add more training data, don't use skip training
+        if self.skip_training and os.path.exists(self.settings_file):
 
-        # If we have training data saved from a previous run of dedupe,
-        # look for it an load it in.
-        # __Note:__ if you want to train from scratch, delete the training_file
+            # Load our deduper from the last training session cache.
+            logging.info('reading from previous training cache %s'
+                                                          % self.settings_file)
+            with open(self.settings_file, 'rb') as f:
+                deduper = dedupe.StaticDedupe(f)
 
-        if os.path.exists(self.training_file):
-            logging.info('reading labeled examples from %s' %
-                         self.training_file)
-            with open(self.training_file) as tf:
-                deduper.readTraining(tf)
-        elif self.skip_training:
-            raise self.parser.error(
-                "You need to provide an existing training_file or run this script without --skip_training")
-
-        if not self.skip_training:
-            logging.info('starting active labeling...')
-
-            dedupe.consoleLabel(deduper)
-
-            # When finished, save our training away to disk
-            logging.info('saving training data to %s' % self.training_file)
-            if sys.version < '3' :
-                with open(self.training_file, 'wb') as tf:
-                    deduper.writeTraining(tf)
-            else :
-                with open(self.training_file, 'w') as tf:
-                    deduper.writeTraining(tf)
+            fields = {variable.field for variable in deduper.data_model.primary_fields}
+            unique_d, parents = exact_matches(data_d, fields)
+                
         else:
-            logging.info('skipping the training step')
+            # # Create a new deduper object and pass our data model to it.
+            deduper = dedupe.Dedupe(self.field_definition)
 
-        deduper.train()
+            fields = {variable.field for variable in deduper.data_model.primary_fields}
+            unique_d, parents = exact_matches(data_d, fields)
+
+            # Set up our data sample
+            logging.info('taking a sample of %d possible pairs', self.sample_size)
+            deduper.sample(unique_d, self.sample_size)
+
+            # Perform standard training procedures
+            self.dedupe_training(deduper)
 
         # ## Blocking
 
@@ -131,13 +124,25 @@ class CSVDedupe(csvhelpers.CSVCommand) :
 
         logging.info('finding a good threshold with a recall_weight of %s' %
                      self.recall_weight)
-        threshold = deduper.threshold(data_d, recall_weight=self.recall_weight)
+        threshold = deduper.threshold(unique_d, recall_weight=self.recall_weight)
 
         # `duplicateClusters` will return sets of record IDs that dedupe
         # believes are all referring to the same entity.
 
         logging.info('clustering...')
-        clustered_dupes = deduper.match(data_d, threshold)
+        clustered_dupes = deduper.match(unique_d, threshold)
+
+        expanded_clustered_dupes = []
+        for cluster, scores in clustered_dupes:
+            new_cluster = list(cluster)
+            new_scores = list(scores)
+            for row_id, score in zip(cluster, scores):
+                children = parents.get(row_id, [])
+                new_cluster.extend(children)
+                new_scores.extend([score] * len(children))
+            expanded_clustered_dupes.append((new_cluster, new_scores))
+
+        clustered_dupes = expanded_clustered_dupes
 
         logging.info('# duplicate sets %s' % len(clustered_dupes))
 
@@ -147,14 +152,27 @@ class CSVDedupe(csvhelpers.CSVCommand) :
             write_function = csvhelpers.writeUniqueResults
 
         if self.output_file:
-            if sys.version < '3' :
-                with open(self.output_file, 'wb') as output_file:
-                    write_function(clustered_dupes, self.input, output_file)
-            else :
-                with open(self.output_file, 'w') as output_file:
-                    write_function(clustered_dupes, self.input, output_file)
+            with open(self.output_file, 'w', encoding='utf-8') as output_file:
+                write_function(clustered_dupes, self.input, output_file)
         else:
-            write_function(clustered_dupes, self.input, sys.stdout)
+            if sys.version < '3' :
+                out = codecs.getwriter(locale.getpreferredencoding())(sys.stdout)
+                write_function(clustered_dupes, self.input, out)
+            else :
+                write_function(clustered_dupes, self.input, sys.stdout)
+
+def exact_matches(data_d, match_fields):
+    unique = {}
+    redundant = {}
+    for key, record in data_d.items():
+        record_hash = hash(tuple(record[f] for f in match_fields))
+        if record_hash not in redundant:
+            unique[key] = record
+            redundant[record_hash] = (key, [])
+        else:
+            redundant[record_hash][1].append(key)
+
+    return unique, {k : v for k, v in redundant.values()}
 
 
 def launch_new_instance():
