@@ -18,13 +18,13 @@ class CSVLink(csvhelpers.CSVCommand):
 
         if len(self.configuration['input']) == 2:
             try:
-                self.input_1 = open(self.configuration['input'][0], 'rU').read()
+                self.input_1 = open(self.configuration['input'][0], encoding='utf-8').read()
             except IOError:
                 raise self.parser.error("Could not find the file %s" %
                                    (self.configuration['input'][0], ))
 
             try:
-                self.input_2 = open(self.configuration['input'][1], 'rU').read()
+                self.input_2 = open(self.configuration['input'][1], encoding='utf-8').read()
             except IOError:
                 raise self.parser.error("Could not find the file %s" %
                                    (self.configuration['input'][1], ))
@@ -71,9 +71,11 @@ class CSVLink(csvhelpers.CSVCommand):
         # import the specified CSV file
 
         data_1 = csvhelpers.readData(self.input_1, self.field_names_1,
-                                     prefix='input_1')
+                                    delimiter=self.delimiter,
+                                    prefix='input_1')
         data_2 = csvhelpers.readData(self.input_2, self.field_names_2,
-                                     prefix='input_2')
+                                    delimiter=self.delimiter,
+                                    prefix='input_2')
 
         # sanity check for provided field names in CSV file
         for field in self.field_names_1:
@@ -99,43 +101,40 @@ class CSVLink(csvhelpers.CSVCommand):
 
         logging.info('using fields: %s' % [field['field']
                                            for field in self.field_definition])
-        # # Create a new deduper object and pass our data model to it.
-        deduper = dedupe.RecordLink(self.field_definition)
 
-        # Set up our data sample
-        logging.info('taking a sample of %d possible pairs', self.sample_size)
-        deduper.sample(data_1, data_2, self.sample_size)
+        # If --skip_training has been selected, and we have a settings cache still
+        # persisting from the last run, use it in this next run.
+        # __Note:__ if you want to add more training data, don't use skip training
+        if self.skip_training and os.path.exists(self.settings_file):
 
-        # If we have training data saved from a previous run of dedupe,
-        # look for it an load it in.
-        # __Note:__ if you want to train from scratch, delete the training_file
+            # Load our deduper from the last training session cache.
+            logging.info('reading from previous training cache %s'
+                                                          % self.settings_file)
+            with open(self.settings_file, 'rb') as f:
+                deduper = dedupe.StaticRecordLink(f)
+                
 
-        if os.path.exists(self.training_file):
-            logging.info('reading labeled examples from %s' %
-                         self.training_file)
-            with open(self.training_file) as tf:
-                deduper.readTraining(tf)
-        elif self.skip_training:
-            raise self.parser.error(
-                "You need to provide an existing training_file or run this script without --skip_training")
-
-        if not self.skip_training:
-            logging.info('starting active labeling...')
-
-            dedupe.consoleLabel(deduper)
-
-            # When finished, save our training away to disk
-            logging.info('saving training data to %s' % self.training_file)
-            if sys.version < '3' :
-                with open(self.training_file, 'wb') as tf:
-                    deduper.writeTraining(tf)
-            else :
-                with open(self.training_file, 'w') as tf:
-                    deduper.writeTraining(tf)
+            fields = {variable.field for variable in deduper.data_model.primary_fields}
+            (nonexact_1,
+             nonexact_2,
+             exact_pairs) = exact_matches(data_1, data_2, fields)
+            
+            
         else:
-            logging.info('skipping the training step')
+            # # Create a new deduper object and pass our data model to it.
+            deduper = dedupe.RecordLink(self.field_definition)
 
-        deduper.train()
+            fields = {variable.field for variable in deduper.data_model.primary_fields}
+            (nonexact_1,
+             nonexact_2,
+             exact_pairs) = exact_matches(data_1, data_2, fields)
+
+            # Set up our data sample
+            logging.info('taking a sample of %d possible pairs', self.sample_size)
+            deduper.sample(nonexact_1, nonexact_2, self.sample_size)
+
+            # Perform standard training procedures
+            self.dedupe_training(deduper)
 
         # ## Blocking
 
@@ -161,6 +160,8 @@ class CSVLink(csvhelpers.CSVCommand):
         logging.info('clustering...')
         clustered_dupes = deduper.match(data_1, data_2, threshold)
 
+        clustered_dupes.extend(exact_pairs)
+
         logging.info('# duplicate sets %s' % len(clustered_dupes))
 
         write_function = csvhelpers.writeLinkedResults
@@ -168,11 +169,11 @@ class CSVLink(csvhelpers.CSVCommand):
 
         if self.output_file:
             if sys.version < '3' :
-                with open(self.output_file, 'wb') as output_file:
+                with open(self.output_file, 'wb', encoding='utf-8') as output_file:
                     write_function(clustered_dupes, self.input_1, self.input_2,
                                    output_file, self.inner_join)
             else :
-                with open(self.output_file, 'w') as output_file:
+                with open(self.output_file, 'w', encoding='utf-8') as output_file:
                     write_function(clustered_dupes, self.input_1, self.input_2,
                                    output_file, self.inner_join)
         else:
@@ -180,6 +181,30 @@ class CSVLink(csvhelpers.CSVCommand):
                            sys.stdout, self.inner_join)
 
 
+def exact_matches(data_1, data_2, match_fields):
+    nonexact_1 = {}
+    nonexact_2 = {}
+    exact_pairs = []
+    redundant = {}
+
+    for key, record in data_1.items():
+        record_hash = hash(tuple(record[f] for f in match_fields))
+        redundant[record_hash] = key        
+
+    for key_2, record in data_2.items():
+        record_hash = hash(tuple(record[f] for f in match_fields))
+        if record_hash in redundant:
+            key_1 = redundant[record_hash]
+            exact_pairs.append(((key_1, key_2), 1.0))
+            del redundant[record_hash]
+        else:
+            nonexact_2[key_2] = record
+
+    for key_1 in redundant.values():
+        nonexact_1[key_1] = data_1[key_1]
+        
+    return nonexact_1, nonexact_2, exact_pairs
+
 def launch_new_instance():
     d = CSVLink()
     d.main()
